@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 2004-2015 Open Source Applications Foundation.
+ * Copyright (c) 2004-2019 Open Source Applications Foundation.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -29,64 +29,61 @@
 #include "iterators.h"
 #include "macros.h"
 
+/* UMemory */
 
-/* UObject */
-
-static PyObject *t_uobject_new(PyTypeObject *type,
+static PyObject *t_umemory_new(PyTypeObject *type,
                                PyObject *args, PyObject *kwds);
 
-static PyObject *t_uobject_richcmp(t_uobject *, PyObject *o2, int op);
-static PyObject *t_uobject_str(t_uobject *self);
-static PyObject *t_uobject_repr(t_uobject *self);
-static PyObject *t_uobject_getDynamicClassID(t_uobject *self);
-static PyObject *t_uobject__getOwned(t_uobject *self, void *data);
+static PyObject *t_umemory_str(t_umemory *self);
+static PyObject *t_umemory_repr(t_umemory *self);
 
-static PyMemberDef t_uobject_members[] = {
-    { NULL, 0, 0, 0, NULL }
-};
-
-static PyMethodDef t_uobject_methods[] = {
-    DECLARE_METHOD(t_uobject, getDynamicClassID, METH_NOARGS),
+static PyMethodDef t_umemory_methods[] = {
     { NULL, NULL, 0, NULL }
 };
 
-static PyGetSetDef t_uobject_properties[] = {
-    { (char *) "owned", (getter) t_uobject__getOwned, NULL, (char *) "", NULL },
+static PyMemberDef t_umemory_members[] = {
+    { NULL, 0, 0, 0, NULL }
+};
+
+static PyObject *t_umemory__getOwned(t_umemory *self, void *data);
+
+static PyGetSetDef t_umemory_properties[] = {
+    { (char *) "owned", (getter) t_umemory__getOwned, NULL, (char *) "", NULL },
     { NULL, NULL, NULL, NULL, NULL }
 };
 
-PyTypeObject UObjectType_ = {
+PyTypeObject UMemoryType_ = {
     PyVarObject_HEAD_INIT(NULL, 0)
-    "icu.UObject",                       /* tp_name */
-    sizeof(t_uobject),                   /* tp_basicsize */
+    "icu.UMemory",                       /* tp_name */
+    sizeof(t_umemory),                   /* tp_basicsize */
     0,                                   /* tp_itemsize */
-    (destructor)t_uobject_dealloc,       /* tp_dealloc */
+    (destructor)t_umemory_dealloc,       /* tp_dealloc */
     0,                                   /* tp_print */
     0,                                   /* tp_getattr */
     0,                                   /* tp_setattr */
     0,                                   /* tp_compare */
-    (reprfunc)t_uobject_repr,            /* tp_repr */
+    (reprfunc)t_umemory_repr,            /* tp_repr */
     0,                                   /* tp_as_number */
     0,                                   /* tp_as_sequence */
     0,                                   /* tp_as_mapping */
     0,                                   /* tp_hash  */
     0,                                   /* tp_call */
-    (reprfunc)t_uobject_str,             /* tp_str */
+    (reprfunc)t_umemory_str,             /* tp_str */
     0,                                   /* tp_getattro */
     0,                                   /* tp_setattro */
     0,                                   /* tp_as_buffer */
     (Py_TPFLAGS_DEFAULT |
      Py_TPFLAGS_BASETYPE),               /* tp_flags */
-    "t_uobject objects",                 /* tp_doc */
+    "t_umemory objects",                 /* tp_doc */
     0,                                   /* tp_traverse */
     0,                                   /* tp_clear */
-    (richcmpfunc)t_uobject_richcmp,      /* tp_richcompare */
+    0,                                   /* tp_richcompare */
     0,                                   /* tp_weaklistoffset */
     0,                                   /* tp_iter */
     0,                                   /* tp_iternext */
-    t_uobject_methods,                   /* tp_methods */
-    t_uobject_members,                   /* tp_members */
-    t_uobject_properties,                /* tp_getset */
+    t_umemory_methods,                   /* tp_methods */
+    t_umemory_members,                   /* tp_members */
+    t_umemory_properties,                /* tp_getset */
     0,                                   /* tp_base */
     0,                                   /* tp_dict */
     0,                                   /* tp_descr_get */
@@ -94,34 +91,24 @@ PyTypeObject UObjectType_ = {
     0,                                   /* tp_dictoffset */
     (initproc)abstract_init,             /* tp_init */
     0,                                   /* tp_alloc */
-    (newfunc)t_uobject_new,              /* tp_new */
+    (newfunc)t_umemory_new,              /* tp_new */
 };
 
-PyObject *wrap_UObject(UObject *object, int flags)
-{
-    if (object)
-    {
-        if (ISINSTANCE(object, UnicodeString))
-            return PyUnicode_FromUnicodeString((UnicodeString *) object);
 
-        t_uobject *self = (t_uobject *) UObjectType_.tp_alloc(&UObjectType_, 0);
-        if (self)
-        {
-            self->object = object;
-            self->flags = flags;
-        }
+/* UObject */
 
-        return (PyObject *) self;
-    }
+static PyObject *t_uobject_richcmp(t_uobject *, PyObject *o2, int op);
+static PyObject *t_uobject_str(t_uobject *self);
+static PyObject *t_uobject_getDynamicClassID(t_uobject *self);
 
-    Py_RETURN_NONE;
-}
+static PyMethodDef t_uobject_methods[] = {
+    DECLARE_METHOD(t_uobject, getDynamicClassID, METH_NOARGS),
+    { NULL, NULL, 0, NULL }
+};
+
+DECLARE_TYPE(UObject, t_uobject, UMemory, UObject,
+             abstract_init, t_uobject_dealloc);
 
-static PyObject *t_uobject__getOwned(t_uobject *self, void *data)
-{
-    int b = self->flags & T_OWNED;
-    Py_RETURN_BOOL(b);
-}
 
 /* Replaceable */
 
@@ -146,6 +133,229 @@ static PyMethodDef t_replaceable_methods[] = {
 DECLARE_TYPE(Replaceable, t_replaceable, UObject, Replaceable,
              abstract_init, NULL);
 
+
+#if U_ICU_VERSION_HEX >= VERSION_HEX(55, 0, 0)
+
+/* PythonReplaceable */
+
+class t_python_replaceable : public _wrapper {
+public:
+    PythonReplaceable *object;
+};
+
+static int t_python_replaceable_init(t_python_replaceable *self, PyObject *args, PyObject *kwds);
+static PyObject *t_python_replaceable_extractBetween(t_python_replaceable *self, PyObject *args);
+static PyObject *t_python_replaceable_handleReplaceBetween(t_python_replaceable *self, PyObject *args);
+static PyObject *t_python_replaceable_copy(t_python_replaceable *self, PyObject *args);
+static PyObject *t_python_replaceable_hasMetaData(t_python_replaceable *self);
+
+static PyMethodDef t_python_replaceable_methods[] = {
+    DECLARE_METHOD(t_python_replaceable, hasMetaData, METH_NOARGS),
+    DECLARE_METHOD(t_python_replaceable, extractBetween, METH_VARARGS),
+    DECLARE_METHOD(t_python_replaceable, handleReplaceBetween, METH_VARARGS),
+    DECLARE_METHOD(t_python_replaceable, copy, METH_VARARGS),
+    { NULL, NULL, 0, NULL }
+};
+
+DECLARE_TYPE(PythonReplaceable, t_python_replaceable, Replaceable,
+             PythonReplaceable, t_python_replaceable_init, NULL);
+
+
+PythonReplaceable::PythonReplaceable(PyObject *self) : self_(self)
+{
+    Py_INCREF(self_);
+}
+
+PythonReplaceable::~PythonReplaceable()
+{
+    Py_DECREF(self_);
+}
+
+int32_t PythonReplaceable::getLength() const
+{
+    PyObject *result = PyObject_CallMethod(self_, (char *) "getLength", NULL);
+
+    if (result == NULL)
+        return -1;
+
+#if PY_MAJOR_VERSION < 3
+    if (PyInt_Check(result))
+    {
+        int n = PyInt_AsLong(result);
+        Py_DECREF(result);
+
+        if (!PyErr_Occurred())
+            return n;
+
+        return -1;
+    }
+#endif
+
+    if (PyLong_Check(result))
+    {
+        int n = PyLong_AsLong(result);
+        Py_DECREF(result);
+
+        if (!PyErr_Occurred())
+            return n;
+
+        return -1;
+    }
+
+    PyErr_SetObject(PyExc_TypeError, result);
+    Py_DECREF(result);
+
+    return -1;
+}
+
+#if U_ICU_VERSION_HEX < VERSION_HEX(59, 1, 0)
+  UChar PythonReplaceable::getCharAt(int32_t offset) const
+#else
+  char16_t PythonReplaceable::getCharAt(int32_t offset) const
+#endif
+{
+    PyObject *result = PyObject_CallMethod(
+        self_, (char *) "getCharAt", (char *) "i", offset);
+
+    if (result == NULL)
+        return -1;
+
+#if PY_MAJOR_VERSION < 3
+    if (PyInt_Check(result))
+    {
+        int n = PyInt_AsLong(result);
+        Py_DECREF(result);
+
+        if (PyErr_Occurred())
+            return -1;
+
+#if U_ICU_VERSION_HEX < VERSION_HEX(59, 1, 0)
+        return (UChar) n;
+#else
+        return (char16_t) n;
+#endif
+    }
+#endif
+
+    if (PyLong_Check(result))
+    {
+        int n = PyLong_AsLong(result);
+
+        Py_DECREF(result);
+        if (PyErr_Occurred())
+            return -1;
+
+#if U_ICU_VERSION_HEX < VERSION_HEX(59, 1, 0)
+        return (UChar) n;
+#else
+        return (char16_t) n;
+#endif
+    }
+
+    UnicodeString *u, _u;
+
+    if (!parseArg(result, "S", &u, &_u) && u->length() == 1)
+    {
+        Py_DECREF(result);
+        return u->charAt(0);
+    }
+
+    PyErr_SetObject(PyExc_TypeError, result);
+    Py_DECREF(result);
+
+    return -1;
+}
+
+UChar32 PythonReplaceable::getChar32At(int32_t offset) const
+{
+    PyObject *result = PyObject_CallMethod(
+        self_, (char *) "getChar32At", (char *) "i", offset);
+
+    if (result == NULL)
+        return -1;
+
+#if PY_MAJOR_VERSION < 3
+    if (PyInt_Check(result))
+    {
+        int n = PyInt_AsLong(result);
+        Py_DECREF(result);
+
+        if (PyErr_Occurred())
+            return -1;
+
+        return (UChar32) n;
+    }
+#endif
+
+    if (PyLong_Check(result))
+    {
+        int n = PyLong_AsLong(result);
+
+        Py_DECREF(result);
+        if (PyErr_Occurred())
+            return -1;
+
+        return (UChar32) n;
+    }
+
+    UnicodeString *u, _u;
+
+    if (!parseArg(result, "S", &u, &_u) && u->countChar32() == 1)
+    {
+        Py_DECREF(result);
+        return u->char32At(0);
+    }
+
+    PyErr_SetObject(PyExc_TypeError, result);
+    Py_DECREF(result);
+
+    return -1;
+}
+
+void PythonReplaceable::extractBetween(
+    int32_t start, int32_t limit, UnicodeString &target) const
+{
+    PyObject *result = PyObject_CallMethod(
+        self_, (char *) "extractBetween", (char *) "ii", start, limit);
+    UnicodeString *u, _u;
+
+    if (result != NULL && !parseArg(result, "S", &u, &_u))
+    {
+        target.setTo(*u);
+        Py_DECREF(result);
+    }
+}
+
+void PythonReplaceable::handleReplaceBetween(
+    int32_t start, int32_t limit, const UnicodeString &text)
+{
+    PyObject *obj = PyUnicode_FromUnicodeString(&text);
+    PyObject *result = PyObject_CallMethod(
+        self_, (char *) "handleReplaceBetween", (char *) "iiO",
+        start, limit, obj);
+
+    Py_DECREF(obj);
+    Py_XDECREF(result);
+}
+
+void PythonReplaceable::copy(int32_t start, int32_t limit, int32_t dest)
+{
+    PyObject *result = PyObject_CallMethod(
+        self_, (char *) "copy", (char *) "iii", start, limit, dest);
+    Py_XDECREF(result);
+}
+
+UBool PythonReplaceable::hasMetaData() const
+{
+    PyObject *result = PyObject_CallMethod(self_, (char *) "hasMetaData", NULL);
+    UBool b = PyObject_IsTrue(result);
+    Py_XDECREF(result);
+
+    return b;
+}
+
+#endif
+
 /* UnicodeString */
 
 class t_unicodestring : public _wrapper {
@@ -293,78 +503,6 @@ PyObject *wrap_Formattable(Formattable &formattable)
 }
 
 
-/* MeasureUnit */
-
-class t_measureunit : public _wrapper {
-public:
-    MeasureUnit *object;
-};
-
-static PyMethodDef t_measureunit_methods[] = {
-    { NULL, NULL, 0, NULL }
-};
-
-DECLARE_TYPE(MeasureUnit, t_measureunit, UObject, MeasureUnit,
-             abstract_init, NULL);
-
-/* Measure */
-
-class t_measure : public _wrapper {
-public:
-    Measure *object;
-};
-
-static PyObject *t_measure_getNumber(t_measure *self);
-static PyObject *t_measure_getUnit(t_measure *self);
-
-static PyMethodDef t_measure_methods[] = {
-    DECLARE_METHOD(t_measure, getNumber, METH_NOARGS),
-    DECLARE_METHOD(t_measure, getUnit, METH_NOARGS),
-    { NULL, NULL, 0, NULL }
-};
-
-DECLARE_TYPE(Measure, t_measure, UObject, Measure, abstract_init, NULL);
-
-/* CurrencyUnit */
-
-class t_currencyunit : public _wrapper {
-public:
-    CurrencyUnit *object;
-};
-
-static int t_currencyunit_init(t_currencyunit *self,
-                               PyObject *args, PyObject *kwds);
-static PyObject *t_currencyunit_getISOCurrency(t_currencyunit *self);
-
-static PyMethodDef t_currencyunit_methods[] = {
-    DECLARE_METHOD(t_currencyunit, getISOCurrency, METH_NOARGS),
-    { NULL, NULL, 0, NULL }
-};
-
-DECLARE_TYPE(CurrencyUnit, t_currencyunit, MeasureUnit, CurrencyUnit,
-             t_currencyunit_init, NULL);
-
-/* CurrencyAmount */
-
-class t_currencyamount : public _wrapper {
-public:
-    CurrencyAmount *object;
-};
-
-static int t_currencyamount_init(t_currencyamount *self,
-                                 PyObject *args, PyObject *kwds);
-static PyObject *t_currencyamount_getCurrency(t_currencyamount *self);
-static PyObject *t_currencyamount_getISOCurrency(t_currencyamount *self);
-
-static PyMethodDef t_currencyamount_methods[] = {
-    DECLARE_METHOD(t_currencyamount, getCurrency, METH_NOARGS),
-    DECLARE_METHOD(t_currencyamount, getISOCurrency, METH_NOARGS),
-    { NULL, NULL, 0, NULL }
-};
-
-DECLARE_TYPE(CurrencyAmount, t_currencyamount, Measure, CurrencyAmount,
-             t_currencyamount_init, NULL);
-
 /* StringEnumeration */
 
 class t_stringenumeration : public _wrapper {
@@ -391,6 +529,72 @@ DECLARE_TYPE(StringEnumeration, t_stringenumeration, UObject,
              StringEnumeration, abstract_init, NULL);
 
 
+/* UMemory */
+
+static PyObject *t_umemory_new(PyTypeObject *type,
+                               PyObject *args, PyObject *kwds)
+{
+    t_umemory *self = (t_umemory *) type->tp_alloc(type, 0);
+
+    if (self)
+    {
+        self->object = NULL;
+        self->flags = 0;
+    }
+
+    return (PyObject *) self;
+}
+
+void t_umemory_dealloc(t_umemory *self)
+{
+    if (self->flags & T_OWNED)
+        delete self->object;
+    self->object = NULL;
+
+    Py_TYPE(self)->tp_free((PyObject *) self);
+}
+
+static PyObject *t_umemory__getOwned(t_umemory *self, void *data)
+{
+    int b = self->flags & T_OWNED;
+    Py_RETURN_BOOL(b);
+}
+
+static PyObject *t_umemory_str(t_umemory *self)
+{
+    if (self->object)
+    {
+        char buf[32];
+        sprintf(buf, "0x%llx", (unsigned long long) (intptr_t) self->object);
+
+        return PyString_FromString(buf);
+    }
+
+    return PyString_FromString("<null>");
+}
+
+static PyObject *t_umemory_repr(t_umemory *self)
+{
+    PyObject *name = PyObject_GetAttrString((PyObject *) Py_TYPE(self),
+                                            "__name__");
+    PyObject *str = Py_TYPE(self)->tp_str((PyObject *) self);
+#if PY_VERSION_HEX < 0x02040000
+    PyObject *args = Py_BuildValue("(OO)", name, str);
+#else
+    PyObject *args = PyTuple_Pack(2, name, str);
+#endif
+    PyObject *format = PyString_FromString("<%s: %s>");
+    PyObject *repr = PyString_Format(format, args);
+
+    Py_DECREF(name);
+    Py_DECREF(str);
+    Py_DECREF(args);
+    Py_DECREF(format);
+
+    return repr;
+}
+
+
 /* UObject */
 
 void t_uobject_dealloc(t_uobject *self)
@@ -402,20 +606,6 @@ void t_uobject_dealloc(t_uobject *self)
     Py_TYPE(self)->tp_free((PyObject *) self);
 }
 
-static PyObject *t_uobject_new(PyTypeObject *type,
-                               PyObject *args, PyObject *kwds)
-{
-    t_uobject *self = (t_uobject *) type->tp_alloc(type, 0);
-
-    if (self)
-    {
-        self->object = NULL;
-        self->flags = 0;
-    }
-
-    return (PyObject *) self;
-}
-
 static PyObject *t_uobject_richcmp(t_uobject *self, PyObject *arg, int op)
 {
     int b = 0;
@@ -463,27 +653,6 @@ static PyObject *t_uobject_str(t_uobject *self)
     return PyString_FromString("<null>");
 }
 
-static PyObject *t_uobject_repr(t_uobject *self)
-{
-    PyObject *name = PyObject_GetAttrString((PyObject *) Py_TYPE(self),
-                                            "__name__");
-    PyObject *str = Py_TYPE(self)->tp_str((PyObject *) self);
-#if PY_VERSION_HEX < 0x02040000
-    PyObject *args = Py_BuildValue("(OO)", name, str);
-#else
-    PyObject *args = PyTuple_Pack(2, name, str);
-#endif
-    PyObject *format = PyString_FromString("<%s: %s>");
-    PyObject *repr = PyString_Format(format, args);
-
-    Py_DECREF(name);
-    Py_DECREF(str);
-    Py_DECREF(args);
-    Py_DECREF(format);
-
-    return repr;
-}
-
 static PyObject *t_uobject_getDynamicClassID(t_uobject *self)
 {
     return PyInt_FromLong((long) self->object->getDynamicClassID());
@@ -540,6 +709,102 @@ static PyObject *t_replaceable_hasMetaData(t_replaceable *self)
 }
 
 
+#if U_ICU_VERSION_HEX >= VERSION_HEX(55, 0, 0)
+
+/* PythonReplaceable */
+
+static int t_python_replaceable_init(
+    t_python_replaceable *self, PyObject *args, PyObject *kwds)
+{
+    PyObject *object;
+
+    switch (PyTuple_Size(args)) {
+      case 1:
+        if (!parseArgs(args, "K", &object))
+        {
+            self->object = new PythonReplaceable(object);
+            self->flags = T_OWNED;
+        }
+        else
+            PyErr_SetArgsError((PyObject *) self, "__init__", args);
+        break;
+
+      default:
+        PyErr_SetArgsError((PyObject *) self, "__init__", args);
+        return -1;
+    }
+
+    if (self->object)
+        return 0;
+
+    return -1;
+}
+
+static PyObject *t_python_replaceable_extractBetween(
+    t_python_replaceable *self, PyObject *args)
+{
+    int start, limit;
+
+    switch (PyTuple_Size(args)) {
+      case 2:
+        if (!parseArgs(args, "ii", &start, &limit))
+        {
+            UnicodeString result;
+            self->object->extractBetween(start, limit, result);
+
+            return PyUnicode_FromUnicodeString(&result);
+        }
+        break;
+    }
+
+    return PyErr_SetArgsError((PyObject *) self, "extractBetween", args);
+}
+
+static PyObject *t_python_replaceable_handleReplaceBetween(
+    t_python_replaceable *self, PyObject *args)
+{
+    int start, limit;
+    UnicodeString *u, _u;
+
+    switch (PyTuple_Size(args)) {
+      case 3:
+        if (!parseArgs(args, "iiS", &start, &limit, &u, &_u))
+        {
+            self->object->handleReplaceBetween(start, limit, *u);
+            Py_RETURN_NONE;
+        }
+        break;
+    }
+
+    return PyErr_SetArgsError((PyObject *) self, "handleReplaceBetween", args);
+}
+
+static PyObject *t_python_replaceable_copy(
+    t_python_replaceable *self, PyObject *args)
+{
+    int start, limit, dest;
+
+    switch (PyTuple_Size(args)) {
+      case 3:
+        if (!parseArgs(args, "iii", &start, &limit, &dest))
+        {
+            self->object->copy(start, limit, dest);
+            Py_RETURN_NONE;
+        }
+        break;
+    }
+
+    return PyErr_SetArgsError((PyObject *) self, "copy", args);
+}
+
+static PyObject *t_python_replaceable_hasMetaData(t_python_replaceable *self)
+{
+    int b = self->object->hasMetaData();
+    Py_RETURN_BOOL(b);
+}
+
+#endif
+
 /* UnicodeString */
 
 static int t_unicodestring_init(t_unicodestring *self,
@@ -556,6 +821,7 @@ static int t_unicodestring_init(t_unicodestring *self,
         self->object = new UnicodeString();
         self->flags = T_OWNED;
         break;
+
       case 1:
         if (!parseArgs(args, "u", &u))
         {
@@ -577,6 +843,7 @@ static int t_unicodestring_init(t_unicodestring *self,
         }
         PyErr_SetArgsError((PyObject *) self, "__init__", args);
         return -1;
+
       case 2:
         if (!parseArgs(args, "Cn", &obj, &encoding))
         {
@@ -600,6 +867,7 @@ static int t_unicodestring_init(t_unicodestring *self,
         }
         PyErr_SetArgsError((PyObject *) self, "__init__", args);
         return -1;
+
       case 3:
         if (!parseArgs(args, "Cnn", &obj, &encoding, &mode))
         {
@@ -621,11 +889,12 @@ static int t_unicodestring_init(t_unicodestring *self,
         }
         PyErr_SetArgsError((PyObject *) self, "__init__", args);
         return -1;
+
       default:
         PyErr_SetArgsError((PyObject *) self, "__init__", args);
         return -1;
     }
-        
+
     if (self->object)
         return 0;
 
@@ -660,7 +929,7 @@ static PyObject *t_unicodestring_getAvailableEncodings(PyTypeObject *type,
       default:
         return PyErr_SetArgsError(type, "getAvailableEncodings", args);
     }
-        
+
     int count = ucnv_countAvailable();
     PyObject *list = PyList_New(0);
 
@@ -754,8 +1023,7 @@ static int verifyStartEnd(int &start, int &end, int len)
 
 static PyObject *t_unicodestring_append(t_unicodestring *self, PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int i, start, len;
 
     switch (PyTuple_Size(args)) {
@@ -793,14 +1061,13 @@ static PyObject *t_unicodestring_append(t_unicodestring *self, PyObject *args)
         }
         break;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "append", args);
 }
 
 static PyObject *t_unicodestring_compare(t_unicodestring *self, PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int start, len;
 
     switch (PyTuple_Size(args)) {
@@ -824,7 +1091,7 @@ static PyObject *t_unicodestring_compare(t_unicodestring *self, PyObject *args)
         }
         break;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "compare", args);
 }
 
@@ -844,19 +1111,18 @@ static PyObject *t_unicodestring_compareBetween(t_unicodestring *self,
                                                  srcStart, srcEnd);
             return PyInt_FromLong(c);
         }
-        
+
         PyErr_SetObject(PyExc_IndexError, args);
         return NULL;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "compareBetween", args);
 }
 
 static PyObject *t_unicodestring_compareCodePointOrder(t_unicodestring *self,
                                                        PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int start, len;
 
     switch (PyTuple_Size(args)) {
@@ -881,14 +1147,13 @@ static PyObject *t_unicodestring_compareCodePointOrder(t_unicodestring *self,
         }
         break;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "compareCodePointOrder", args);
 }
 
 static PyObject *t_unicodestring_compareCodePointOrderBetween(t_unicodestring *self, PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int start, end, srcStart, srcEnd;
 
     if (!parseArgs(args, "iiSii", &start, &end, &u, &_u, &srcStart, &srcEnd))
@@ -899,11 +1164,11 @@ static PyObject *t_unicodestring_compareCodePointOrderBetween(t_unicodestring *s
             int c = self->object->compareCodePointOrderBetween(start, end, *u, srcStart, srcEnd);
             return PyInt_FromLong(c);
         }
-        
+
         PyErr_SetObject(PyExc_IndexError, args);
         return NULL;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "compareCodePointOrderBetween",
                               args);
 }
@@ -911,8 +1176,7 @@ static PyObject *t_unicodestring_compareCodePointOrderBetween(t_unicodestring *s
 static PyObject *t_unicodestring_caseCompare(t_unicodestring *self,
                                              PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int start, len, options;
 
     switch (PyTuple_Size(args)) {
@@ -937,15 +1201,14 @@ static PyObject *t_unicodestring_caseCompare(t_unicodestring *self,
         }
         break;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "caseCompare", args);
 }
 
 static PyObject *t_unicodestring_caseCompareBetween(t_unicodestring *self,
                                                     PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int start, end, srcStart, srcEnd, options;
 
     if (!parseArgs(args, "iiSiii", &start, &end, &u, &_u, &srcStart, &srcEnd,
@@ -958,19 +1221,18 @@ static PyObject *t_unicodestring_caseCompareBetween(t_unicodestring *self,
                                                      srcStart, srcEnd, options);
             return PyInt_FromLong(c);
         }
-        
+
         PyErr_SetObject(PyExc_IndexError, args);
         return NULL;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "caseCompareBetween", args);
 }
 
 static PyObject *t_unicodestring_startsWith(t_unicodestring *self,
                                             PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int start, len;
 
     switch (PyTuple_Size(args)) {
@@ -995,15 +1257,14 @@ static PyObject *t_unicodestring_startsWith(t_unicodestring *self,
         }
         break;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "startsWith", args);
 }
 
 static PyObject *t_unicodestring_endsWith(t_unicodestring *self,
                                           PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int start, len;
 
     switch (PyTuple_Size(args)) {
@@ -1028,14 +1289,13 @@ static PyObject *t_unicodestring_endsWith(t_unicodestring *self,
         }
         break;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "endsWith", args);
 }
 
 static PyObject *t_unicodestring_indexOf(t_unicodestring *self, PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int c, start, len, srcStart, srcLen;
 
     switch (PyTuple_Size(args)) {
@@ -1134,8 +1394,7 @@ static PyObject *t_unicodestring_indexOf(t_unicodestring *self, PyObject *args)
 static PyObject *t_unicodestring_lastIndexOf(t_unicodestring *self,
                                              PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     int c, start, len, srcStart, srcLen;
 
     switch (PyTuple_Size(args)) {
@@ -1363,7 +1622,7 @@ static PyObject *t_unicodestring_repr(t_unicodestring *self)
     }
     if (!str)
         return NULL;
-    
+
 #if PY_VERSION_HEX < 0x02040000
     PyObject *args = Py_BuildValue("(OO)", name, str);
 #else
@@ -1382,7 +1641,7 @@ static PyObject *t_unicodestring_repr(t_unicodestring *self)
 
 static long t_unicodestring_hash(t_unicodestring *self)
 {
-  return (long) self->object->hashCode();
+    return (long) self->object->hashCode();
 }
 
 
@@ -1664,7 +1923,7 @@ static PyObject *t_unicodestring_richcmp(t_unicodestring *self,
         } catch (ICUException e) {
             return e.reportError();
         }
-        
+
     switch (op) {
       case Py_EQ:
         b = *self->object == *u;
@@ -1692,6 +1951,7 @@ static PyObject *t_unicodestring_richcmp(t_unicodestring *self,
     Py_RETURN_BOOL(b);
 }
 
+// unicodestring as sequence of UChar, not codepoints
 static Py_ssize_t t_unicodestring_length(t_unicodestring *self)
 {
     return self->object->length();
@@ -1714,10 +1974,7 @@ static PyObject *t_unicodestring_concat(t_unicodestring *self, PyObject *arg)
     {
         UnicodeString *v = new UnicodeString(*self->object);
 
-        if (sizeof(Py_UNICODE) == sizeof(UChar))
-            v->append((UChar) i);
-        else
-            v->append((UChar32) i);
+        v->append((UChar32) i);
 
         return wrap_UnicodeString(v, T_OWNED);
     }
@@ -1751,16 +2008,9 @@ static PyObject *t_unicodestring_item(t_unicodestring *self, int n)
 
     if (n >= 0 && n < len)
     {
-        if (sizeof(Py_UNICODE) == sizeof(UChar))
-        {
-            Py_UNICODE c = (Py_UNICODE) u->charAt(n);
-            return PyUnicode_FromUnicode(&c, 1);
-        }
-        else
-        {
-            Py_UNICODE c = (Py_UNICODE) u->char32At(n);
-            return PyUnicode_FromUnicode(&c, 1);
-        }
+        // unicodestring as sequence of UChar, not codepoints
+        Py_UNICODE c = (Py_UNICODE) u->charAt(n);
+        return PyUnicode_FromUnicode(&c, 1);
     }
 
     PyErr_SetNone(PyExc_IndexError);
@@ -1788,7 +2038,7 @@ static PyObject *t_unicodestring_slice(t_unicodestring *self,
     if (l >= 0 && h >= 0)
     {
         if (h > l)
-          u->extract((int32_t) l, (int32_t) (h - l), *v);
+            u->extract((int32_t) l, (int32_t) (h - l), *v);
 
         return wrap_UnicodeString(v, T_OWNED);
     }
@@ -1812,11 +2062,8 @@ static int t_unicodestring_ass_item(t_unicodestring *self,
 
         if (!parseArg(arg, "i", &i))
         {
-            if (sizeof(Py_UNICODE) == sizeof(UChar))
-                u->replace((int32_t) n, 1, (UChar) i);
-            else
-                u->replace((int32_t) n, 1, (UChar32) i);
-
+            // unicodestring as sequence of UChar, not codepoints
+            u->replace((int32_t) n, 1, (UChar) i);
             return 0;
         }
 
@@ -1827,6 +2074,7 @@ static int t_unicodestring_ass_item(t_unicodestring *self,
         {
             if (v->length() == 1)
             {
+                // unicodestring as sequence of UChar, not codepoints
                 u->setCharAt((int32_t) n, v->charAt(0));
                 return 0;
             }
@@ -1913,10 +2161,7 @@ static PyObject *t_unicodestring_inplace_concat(t_unicodestring *self,
 
     if (!parseArg(arg, "i", &i))
     {
-        if (sizeof(Py_UNICODE) == sizeof(UChar))
-            self->object->append((UChar) i);
-        else
-            self->object->append((UChar32) i);
+        self->object->append((UChar32) i);
 
         Py_INCREF(self);
         return (PyObject *) self;
@@ -1941,6 +2186,7 @@ static PyObject *t_unicodestring_inplace_repeat(t_unicodestring *self,
     return (PyObject *) self;
 }
 
+// unicodestring as sequence of UChar, not codepoints
 static PySequenceMethods t_unicodestring_as_sequence = {
     (lenfunc) t_unicodestring_length,                   /* sq_length */
     (binaryfunc) t_unicodestring_concat,                /* sq_concat */
@@ -2220,32 +2466,7 @@ static PyObject *t_formattable_setString(t_formattable *self, PyObject *arg)
     return PyErr_SetArgsError((PyObject *) self, "setString", arg);
 }
 
-static PyObject *t_formattable_richcmp(t_formattable *self,
-                                       PyObject *arg, int op)
-{
-    Formattable *f;
-
-    if (!parseArg(arg, "P", TYPE_CLASSID(Formattable), &f))
-    {
-        int b = 0;
-
-        switch (op) {
-          case Py_EQ:
-            b = *self->object == *f;
-            break;
-          case Py_NE:
-            b = *self->object != *f;
-            break;
-          default:
-            PyErr_SetNone(PyExc_NotImplementedError);
-            return NULL;
-        }
-
-        Py_RETURN_BOOL(b);
-    }
-
-    return PyErr_SetArgsError((PyObject *) self, "__richcmp__", arg);
-}
+DEFINE_RICHCMP(Formattable, t_formattable);
 
 static PyObject *t_formattable_str(t_formattable *self)
 {
@@ -2305,7 +2526,7 @@ static PyObject *t_formattable_repr(t_formattable *self)
     }
     if (!str)
         return NULL;
-    
+
 #if PY_VERSION_HEX < 0x02040000
     PyObject *args = Py_BuildValue("(OO)", name, str);
 #else
@@ -2323,204 +2544,13 @@ static PyObject *t_formattable_repr(t_formattable *self)
 }
 
 
-/* MeasureUnit */
-
-static PyObject *t_measureunit_richcmp(t_measureunit *self,
-                                       PyObject *arg, int op)
-{
-    int b = 0;
-
-    switch (op) {
-      case Py_EQ:
-      case Py_NE:
-        if (PyObject_TypeCheck(arg, &UObjectType_))
-            b = *self->object == *((t_uobject *) arg)->object;
-        if (op == Py_EQ)
-            Py_RETURN_BOOL(b);
-        Py_RETURN_BOOL(!b);
-      case Py_LT:
-      case Py_LE:
-      case Py_GT:
-      case Py_GE:
-        PyErr_SetNone(PyExc_NotImplementedError);
-        return NULL;
-    }
-
-    return NULL;
-}
-
-
-/* Measure */
-
-static PyObject *t_measure_getNumber(t_measure *self)
-{
-    Formattable *f = new Formattable(self->object->getNumber());
-    return wrap_Formattable(f, T_OWNED);
-}
-
-static PyObject *t_measure_getUnit(t_measure *self)
-{
-    MeasureUnit *u = (MeasureUnit *) self->object->getUnit().clone();
-    return wrap_MeasureUnit(u, T_OWNED);
-}
-
-static PyObject *t_measure_richcmp(t_measure *self, PyObject *arg, int op)
-{
-    int b = 0;
-
-    switch (op) {
-      case Py_EQ:
-      case Py_NE:
-        if (PyObject_TypeCheck(arg, &UObjectType_))
-            b = *self->object == *((t_uobject *) arg)->object;
-        if (op == Py_EQ)
-            Py_RETURN_BOOL(b);
-        Py_RETURN_BOOL(!b);
-      case Py_LT:
-      case Py_LE:
-      case Py_GT:
-      case Py_GE:
-        PyErr_SetNone(PyExc_NotImplementedError);
-        return NULL;
-    }
-
-    return NULL;
-}
-
-
-/* CurrencyUnit */
-
-static int t_currencyunit_init(t_currencyunit *self,
-                               PyObject *args, PyObject *kwds)
-{
-    UErrorCode status = U_ZERO_ERROR;
-    UnicodeString *u;
-    UnicodeString _u;
-
-    if (!parseArgs(args, "S", &u, &_u))
-    {
-        CurrencyUnit *cu = new CurrencyUnit(u->getTerminatedBuffer(), status);
-
-        if (U_FAILURE(status))
-        {
-            ICUException(status).reportError();
-            return -1;
-        }
-
-        self->object = cu;
-        self->flags = T_OWNED;
-
-        return 0;
-    }
-
-    PyErr_SetArgsError((PyObject *) self, "__init__", args);
-    return -1;
-}
-
-static PyObject *t_currencyunit_getISOCurrency(t_currencyunit *self)
-{
-    UnicodeString u(self->object->getISOCurrency());
-    return PyUnicode_FromUnicodeString(&u);
-}
-
-static PyObject *t_currencyunit_str(t_currencyunit *self)
-{
-    UnicodeString u(self->object->getISOCurrency());
-    return PyUnicode_FromUnicodeString(&u);
-}
-
-
-/* CurrencyAmount */
-
-static int t_currencyamount_init(t_currencyamount *self,
-                                 PyObject *args, PyObject *kwds)
-{
-    UErrorCode status = U_ZERO_ERROR;
-    Formattable *f;
-    double d;
-    UnicodeString *u;
-    UnicodeString _u;
-
-    if (!parseArgs(args, "PS", TYPE_CLASSID(Formattable),
-                   &f, &u, &_u))
-    {
-        CurrencyAmount *ca =
-            new CurrencyAmount(*f, u->getTerminatedBuffer(), status);
-
-        if (U_FAILURE(status))
-        {
-            ICUException(status).reportError();
-            return -1;
-        }
-
-        self->object = ca;
-        self->flags = T_OWNED;
-
-        return 0;
-    }
-
-    if (!parseArgs(args, "dS", &d, &u, &_u))
-    {
-        CurrencyAmount *ca =
-            new CurrencyAmount(d, u->getTerminatedBuffer(), status);
-
-        if (U_FAILURE(status))
-        {
-            ICUException(status).reportError();
-            return -1;
-        }
-
-        self->object = ca;
-        self->flags = T_OWNED;
-
-        return 0;
-    }
-
-    PyErr_SetArgsError((PyObject *) self, "__init__", args);
-    return -1;
-}
-
-static PyObject *t_currencyamount_getCurrency(t_currencyamount *self)
-{
-    CurrencyUnit *cu = new CurrencyUnit(self->object->getCurrency());
-    return wrap_CurrencyUnit(cu, T_OWNED);
-}
-
-static PyObject *t_currencyamount_getISOCurrency(t_currencyamount *self)
-{
-    UnicodeString u(self->object->getISOCurrency());
-    return PyUnicode_FromUnicodeString(&u);
-}
-
-static PyObject *t_currencyamount_str(t_currencyamount *self)
-{
-    UnicodeString u(self->object->getISOCurrency());
-    UErrorCode status = U_ZERO_ERROR;
-    double d = self->object->getNumber().getDouble(status);
-
-    PyObject *currency = PyUnicode_FromUnicodeString(&u);
-    PyObject *amount = PyFloat_FromDouble(d);
-    PyObject *format = PyString_FromString("%s %0.2f");
-    PyObject *tuple = PyTuple_New(2);
-    PyObject *str;
-
-    PyTuple_SET_ITEM(tuple, 0, currency);
-    PyTuple_SET_ITEM(tuple, 1, amount);
-    str = PyString_Format(format, tuple);
-    Py_DECREF(tuple);
-    Py_DECREF(format);
-
-    return str;
-}
-
-
 /* StringEnumeration */
 
 static PyObject *t_stringenumeration_count(t_stringenumeration *self)
 {
     UErrorCode status = U_ZERO_ERROR;
     int i = self->object->count(status);
-  
+
     if (U_FAILURE(status))
         return ICUException(status).reportError();
 
@@ -2530,7 +2560,7 @@ static PyObject *t_stringenumeration_count(t_stringenumeration *self)
 static PyObject *t_stringenumeration_reset(t_stringenumeration *self)
 {
     UErrorCode status = U_ZERO_ERROR;
-  
+
     self->object->reset(status);
     if (U_FAILURE(status))
         return ICUException(status).reportError();
@@ -2601,6 +2631,8 @@ static PyObject *t_stringenumeration_iter(t_stringenumeration *self)
 
 void _init_bases(PyObject *m)
 {
+    UObjectType_.tp_str = (reprfunc) t_uobject_str;
+    UObjectType_.tp_richcompare = (richcmpfunc) t_uobject_richcmp;
     UnicodeStringType_.tp_str = (reprfunc) t_unicodestring_str;
     UnicodeStringType_.tp_repr = (reprfunc) t_unicodestring_repr;
     UnicodeStringType_.tp_richcompare = (richcmpfunc) t_unicodestring_richcmp;
@@ -2612,26 +2644,34 @@ void _init_bases(PyObject *m)
     FormattableType_.tp_richcompare = (richcmpfunc) t_formattable_richcmp;
     FormattableType_.tp_str = (reprfunc) t_formattable_str;
     FormattableType_.tp_repr = (reprfunc) t_formattable_repr;
-    MeasureUnitType_.tp_richcompare = (richcmpfunc) t_measureunit_richcmp;
-    MeasureType_.tp_richcompare = (richcmpfunc) t_measure_richcmp;
-    CurrencyUnitType_.tp_str = (reprfunc) t_currencyunit_str;
-    CurrencyAmountType_.tp_str = (reprfunc) t_currencyamount_str;
     StringEnumerationType_.tp_iter = (getiterfunc) t_stringenumeration_iter;
     StringEnumerationType_.tp_iternext = (iternextfunc) t_stringenumeration_next;
 
+    INSTALL_STRUCT(UMemory, m);  // no typeid()
     INSTALL_TYPE(UObject, m);
     INSTALL_TYPE(Replaceable, m);
+#if U_ICU_VERSION_HEX >= VERSION_HEX(55, 0, 0)
+    INSTALL_TYPE(PythonReplaceable, m);
+#endif
     REGISTER_TYPE(UnicodeString, m);
     REGISTER_TYPE(Formattable, m);
-    INSTALL_TYPE(MeasureUnit, m);
-    INSTALL_TYPE(Measure, m);
-    REGISTER_TYPE(CurrencyUnit, m);
-    REGISTER_TYPE(CurrencyAmount, m);
     INSTALL_TYPE(StringEnumeration, m);
 
     INSTALL_MODULE_INT(m, U_FOLD_CASE_DEFAULT);
     INSTALL_MODULE_INT(m, U_COMPARE_CODE_POINT_ORDER);
+    INSTALL_MODULE_INT(m, U_COMPARE_IGNORE_CASE);
     INSTALL_MODULE_INT(m, U_FOLD_CASE_EXCLUDE_SPECIAL_I);
+    INSTALL_MODULE_INT(m, U_TITLECASE_NO_BREAK_ADJUSTMENT);
+    INSTALL_MODULE_INT(m, U_TITLECASE_NO_LOWERCASE);
+    INSTALL_MODULE_INT(m, UNORM_INPUT_IS_FCD);
+
+#if U_ICU_VERSION_HEX >= VERSION_HEX(60, 0, 0)
+    INSTALL_MODULE_INT(m, U_EDITS_NO_RESET);
+    INSTALL_MODULE_INT(m, U_OMIT_UNCHANGED_TEXT);
+    INSTALL_MODULE_INT(m, U_TITLECASE_ADJUST_TO_CASED);
+    INSTALL_MODULE_INT(m, U_TITLECASE_SENTENCES);
+    INSTALL_MODULE_INT(m, U_TITLECASE_WHOLE_STRING);
+#endif
 
 #if U_ICU_VERSION_HEX < VERSION_HEX(55, 1, 0)
     INSTALL_MODULE_INT(m, UIDNA_DEFAULT);
