@@ -30,6 +30,7 @@
 #include "bases.h"
 #include "char.h"
 #include "macros.h"
+#include "unicodeset.h"
 
 DECLARE_CONSTANTS_TYPE(UProperty);
 DECLARE_CONSTANTS_TYPE(UCharDirection);
@@ -40,9 +41,16 @@ DECLARE_CONSTANTS_TYPE(UPropertyNameChoice);
 DECLARE_CONSTANTS_TYPE(UWordBreakValues);
 DECLARE_CONSTANTS_TYPE(UJoiningGroup);
 DECLARE_CONSTANTS_TYPE(ULineBreak);
+DECLARE_CONSTANTS_TYPE(UGraphemeClusterBreak);
+DECLARE_CONSTANTS_TYPE(UHangulSyllableType);
 #if U_ICU_VERSION_HEX >= VERSION_HEX(52, 0, 0)
 DECLARE_CONSTANTS_TYPE(UBidiPairedBracketType);
 #endif
+#if U_ICU_VERSION_HEX >= VERSION_HEX(63, 0, 0)
+DECLARE_CONSTANTS_TYPE(UIndicPositionalCategory);
+DECLARE_CONSTANTS_TYPE(UIndicSyllabicCategory);
+DECLARE_CONSTANTS_TYPE(UVerticalOrientation);
+#endif
 
 
 /* Char */
@@ -57,6 +65,9 @@ typedef UChar32 (*uchar32_char_fn)(UChar32 c);
 
 static int t_char_init(t_char *self, PyObject *args, PyObject *kwds);
 static PyObject *t_char_hasBinaryProperty(PyTypeObject *type, PyObject *args);
+#if U_ICU_VERSION_HEX >= VERSION_HEX(63, 0, 0)
+static PyObject *t_char_getBinaryPropertySet(PyTypeObject *type, PyObject *arg);
+#endif
 static PyObject *t_char_getIntPropertyValue(PyTypeObject *type, PyObject *args);
 static PyObject *t_char_getIntPropertyMinValue(PyTypeObject *type,
                                                PyObject *arg);
@@ -101,7 +112,12 @@ static PyObject *t_char_charFromName(PyTypeObject *type, PyObject *args);
 static PyObject *t_char_enumCharNames(PyTypeObject *type, PyObject *args);
 static PyObject *t_char_getPropertyName(PyTypeObject *type, PyObject *args);
 static PyObject *t_char_getPropertyEnum(PyTypeObject *type, PyObject *arg);
+static PyObject *t_char_getPropertyValueName(PyTypeObject *type,
+                                             PyObject *args);
+static PyObject *t_char_getPropertyValueEnum(PyTypeObject *type,
+                                             PyObject *args);
 static PyObject *t_char_isIDStart(PyTypeObject *type, PyObject *arg);
+static PyObject *t_char_isIDPart(PyTypeObject *type, PyObject *arg);
 static PyObject *t_char_isIDIgnorable(PyTypeObject *type, PyObject *arg);
 static PyObject *t_char_isJavaIDStart(PyTypeObject *type, PyObject *arg);
 static PyObject *t_char_isJavaIDPart(PyTypeObject *type, PyObject *arg);
@@ -117,6 +133,9 @@ static PyObject *t_char_getFC_NFKC_Closure(PyTypeObject *type, PyObject *arg);
 
 static PyMethodDef t_char_methods[] = {
     DECLARE_METHOD(t_char, hasBinaryProperty, METH_VARARGS | METH_CLASS),
+#if U_ICU_VERSION_HEX >= VERSION_HEX(63, 0, 0)
+    DECLARE_METHOD(t_char, getBinaryPropertySet, METH_O | METH_CLASS),
+#endif
     DECLARE_METHOD(t_char, getIntPropertyValue, METH_VARARGS | METH_CLASS),
     DECLARE_METHOD(t_char, getIntPropertyMinValue, METH_O | METH_CLASS),
     DECLARE_METHOD(t_char, getIntPropertyMaxValue, METH_O | METH_CLASS),
@@ -159,7 +178,10 @@ static PyMethodDef t_char_methods[] = {
     DECLARE_METHOD(t_char, enumCharNames, METH_VARARGS | METH_CLASS),
     DECLARE_METHOD(t_char, getPropertyName, METH_VARARGS | METH_CLASS),
     DECLARE_METHOD(t_char, getPropertyEnum, METH_O | METH_CLASS),
+    DECLARE_METHOD(t_char, getPropertyValueName, METH_VARARGS | METH_CLASS),
+    DECLARE_METHOD(t_char, getPropertyValueEnum, METH_VARARGS | METH_CLASS),
     DECLARE_METHOD(t_char, isIDStart, METH_O | METH_CLASS),
+    DECLARE_METHOD(t_char, isIDPart, METH_O | METH_CLASS),
     DECLARE_METHOD(t_char, isIDIgnorable, METH_O | METH_CLASS),
     DECLARE_METHOD(t_char, isJavaIDStart, METH_O | METH_CLASS),
     DECLARE_METHOD(t_char, isJavaIDPart, METH_O | METH_CLASS),
@@ -215,6 +237,24 @@ static PyObject *t_char_hasBinaryProperty(PyTypeObject *type, PyObject *args)
     return PyErr_SetArgsError((PyObject *) type, "hasBinaryProperty", args);
 }
 
+#if U_ICU_VERSION_HEX >= VERSION_HEX(63, 0, 0)
+static PyObject *t_char_getBinaryPropertySet(PyTypeObject *type, PyObject *arg)
+{
+    UProperty prop;
+
+    if (!parseArg(arg, "i", &prop))
+    {
+        const USet *set;
+        STATUS_CALL(set = u_getBinaryPropertySet(prop, &status));
+
+        return wrap_UnicodeSet(
+            const_cast<UnicodeSet *>(UnicodeSet::fromUSet(set)), 0);
+    }
+
+    return PyErr_SetArgsError((PyObject *) type, "getBinaryPropertySet", arg);
+}
+#endif
+
 static PyObject *t_char_getIntPropertyValue(PyTypeObject *type, PyObject *args)
 {
     UnicodeString *u, _u;
@@ -384,6 +424,10 @@ static PyObject *t_char_isIDStart(PyTypeObject *type, PyObject *arg) {
     return t_char_fn(u_isIDStart, "isIDStart", type, arg);
 }
 
+static PyObject *t_char_isIDPart(PyTypeObject *type, PyObject *arg) {
+    return t_char_fn(u_isIDPart, "isIDPart", type, arg);
+}
+
 static PyObject *t_char_isIDIgnorable(PyTypeObject *type, PyObject *arg) {
     return t_char_fn(u_isIDIgnorable, "isIDIgnorable", type, arg);
 }
@@ -664,14 +708,60 @@ static PyObject *t_char_getPropertyName(PyTypeObject *type, PyObject *args)
 
 static PyObject *t_char_getPropertyEnum(PyTypeObject *type, PyObject *arg)
 {
-    char *alias;
+    charsArg alias;
 
-    if (!parseArg(arg, "c", &alias))
+    if (!parseArg(arg, "n", &alias))
         return PyInt_FromLong(u_getPropertyEnum(alias));
 
     return PyErr_SetArgsError((PyObject *) type, "getPropertyEnum", arg);
 }
 
+static PyObject *t_char_getPropertyValueName(PyTypeObject *type, PyObject *args)
+{
+    UPropertyNameChoice choice = U_SHORT_PROPERTY_NAME;
+    int32_t value;
+    UProperty prop;
+    const char *result;
+
+    switch (PyTuple_Size(args)) {
+      case 2:
+        if (!parseArgs(args, "ii", &prop, &value))
+        {
+            result = u_getPropertyValueName(prop, value, choice);
+            if (result != NULL)
+                return PyString_FromString(result);
+            Py_RETURN_NONE;
+        }
+        break;
+      case 3:
+        if (!parseArgs(args, "iii", &prop, &value, &choice))
+        {
+            result = u_getPropertyValueName(prop, value, choice);
+            if (result != NULL)
+                return PyString_FromString(result);
+            Py_RETURN_NONE;
+        }
+        break;
+    }
+
+    return PyErr_SetArgsError((PyObject *) type, "getPropertyValueName", args);
+}
+
+static PyObject *t_char_getPropertyValueEnum(PyTypeObject *type, PyObject *args)
+{
+    UProperty prop;
+    charsArg alias;
+
+    switch (PyTuple_Size(args)) {
+      case 2:
+        if (!parseArgs(args, "in", &prop, &alias))
+            return PyInt_FromLong(u_getPropertyValueEnum(prop, alias));
+        break;
+    }
+
+    return PyErr_SetArgsError((PyObject *) type, "getPropertyValueEnum", args);
+}
+
 static PyObject *t_char_fn(uchar32_char_fn fn, const char *name,
                            PyTypeObject *type, PyObject *arg)
 {
@@ -858,8 +948,15 @@ void _init_char(PyObject *m)
     INSTALL_CONSTANTS_TYPE(UWordBreakValues, m);
     INSTALL_CONSTANTS_TYPE(UJoiningGroup, m);
     INSTALL_CONSTANTS_TYPE(ULineBreak, m);
+    INSTALL_CONSTANTS_TYPE(UGraphemeClusterBreak, m);
+    INSTALL_CONSTANTS_TYPE(UHangulSyllableType, m);
 #if U_ICU_VERSION_HEX >= VERSION_HEX(52, 0, 0)
     INSTALL_CONSTANTS_TYPE(UBidiPairedBracketType, m);
+#endif
+#if U_ICU_VERSION_HEX >= VERSION_HEX(63, 0, 0)
+    INSTALL_CONSTANTS_TYPE(UIndicPositionalCategory, m);
+    INSTALL_CONSTANTS_TYPE(UIndicSyllabicCategory, m);
+    INSTALL_CONSTANTS_TYPE(UVerticalOrientation, m);
 #endif
     INSTALL_STRUCT(Char, m);
 
@@ -986,6 +1083,11 @@ void _init_char(PyObject *m)
 #endif
 #if U_ICU_VERSION_HEX >= VERSION_HEX(62, 0, 0)
     INSTALL_ENUM(UProperty, "EXTENDED_PICTOGRAPHIC", UCHAR_EXTENDED_PICTOGRAPHIC);
+#endif
+#if U_ICU_VERSION_HEX >= VERSION_HEX(63, 0, 0)
+    INSTALL_ENUM(UProperty, "INDIC_POSITIONAL_CATEGORY", UCHAR_INDIC_POSITIONAL_CATEGORY);
+    INSTALL_ENUM(UProperty, "INDIC_SYLLABIC_CATEGORY", UCHAR_INDIC_SYLLABIC_CATEGORY);
+    INSTALL_ENUM(UProperty, "VERTICAL_ORIENTATION", UCHAR_VERTICAL_ORIENTATION);
 #endif
     INSTALL_ENUM(UProperty, "INVALID_CODE ", UCHAR_INVALID_CODE );
 
@@ -1358,6 +1460,17 @@ void _init_char(PyObject *m)
     INSTALL_ENUM(UBlockCode, "MEDEFAIDRIN", UBLOCK_MEDEFAIDRIN);
     INSTALL_ENUM(UBlockCode, "OLD_SOGDIAN", UBLOCK_OLD_SOGDIAN);
     INSTALL_ENUM(UBlockCode, "SOGDIAN", UBLOCK_SOGDIAN);
+#endif
+#if U_ICU_VERSION_HEX >= VERSION_HEX(64, 0, 0)
+    INSTALL_ENUM(UBlockCode, "EGYPTIAN_HIEROGLYPH_FORMAT_CONTROLS", UBLOCK_EGYPTIAN_HIEROGLYPH_FORMAT_CONTROLS);
+    INSTALL_ENUM(UBlockCode, "ELYMAIC", UBLOCK_ELYMAIC);
+    INSTALL_ENUM(UBlockCode, "NANDINAGARI", UBLOCK_NANDINAGARI);
+    INSTALL_ENUM(UBlockCode, "NYIAKENG_PUACHUE_HMONG", UBLOCK_NYIAKENG_PUACHUE_HMONG);
+    INSTALL_ENUM(UBlockCode, "OTTOMAN_SIYAQ_NUMBERS", UBLOCK_OTTOMAN_SIYAQ_NUMBERS);
+    INSTALL_ENUM(UBlockCode, "SMALL_KANA_EXTENSION", UBLOCK_SMALL_KANA_EXTENSION);
+    INSTALL_ENUM(UBlockCode, "SYMBOLS_AND_PICTOGRAPHS_EXTENDED_A", UBLOCK_SYMBOLS_AND_PICTOGRAPHS_EXTENDED_A);
+    INSTALL_ENUM(UBlockCode, "TAMIL_SUPPLEMENT", UBLOCK_TAMIL_SUPPLEMENT);
+    INSTALL_ENUM(UBlockCode, "WANCHO", UBLOCK_WANCHO);
 #endif
     INSTALL_ENUM(UBlockCode, "INVALID_CODE", UBLOCK_INVALID_CODE);
 
@@ -1376,6 +1489,66 @@ void _init_char(PyObject *m)
     INSTALL_ENUM(UBidiPairedBracketType, "CLOSE", U_BPT_CLOSE);
 #endif
 
+#if U_ICU_VERSION_HEX >= VERSION_HEX(63, 0, 0)
+    INSTALL_ENUM(UIndicPositionalCategory, "NA", U_INPC_NA);
+    INSTALL_ENUM(UIndicPositionalCategory, "BOTTOM", U_INPC_BOTTOM);
+    INSTALL_ENUM(UIndicPositionalCategory, "BOTTOM_AND_LEFT", U_INPC_BOTTOM_AND_LEFT);
+    INSTALL_ENUM(UIndicPositionalCategory, "BOTTOM_AND_RIGHT", U_INPC_BOTTOM_AND_RIGHT);
+    INSTALL_ENUM(UIndicPositionalCategory, "LEFT", U_INPC_LEFT);
+    INSTALL_ENUM(UIndicPositionalCategory, "LEFT_AND_RIGHT", U_INPC_LEFT_AND_RIGHT);
+    INSTALL_ENUM(UIndicPositionalCategory, "OVERSTRUCK", U_INPC_OVERSTRUCK);
+    INSTALL_ENUM(UIndicPositionalCategory, "RIGHT", U_INPC_RIGHT);
+    INSTALL_ENUM(UIndicPositionalCategory, "TOP", U_INPC_TOP);
+    INSTALL_ENUM(UIndicPositionalCategory, "TOP_AND_BOTTOM", U_INPC_TOP_AND_BOTTOM);
+    INSTALL_ENUM(UIndicPositionalCategory, "TOP_AND_BOTTOM_AND_RIGHT", U_INPC_TOP_AND_BOTTOM_AND_RIGHT);
+    INSTALL_ENUM(UIndicPositionalCategory, "TOP_AND_LEFT", U_INPC_TOP_AND_LEFT);
+    INSTALL_ENUM(UIndicPositionalCategory, "TOP_AND_LEFT_AND_RIGHT", U_INPC_TOP_AND_LEFT_AND_RIGHT);
+    INSTALL_ENUM(UIndicPositionalCategory, "TOP_AND_RIGHT", U_INPC_TOP_AND_RIGHT);
+    INSTALL_ENUM(UIndicPositionalCategory, "VISUAL_ORDER_LEFT", U_INPC_VISUAL_ORDER_LEFT);
+
+    INSTALL_ENUM(UIndicSyllabicCategory, "OTHER", U_INSC_OTHER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "AVAGRAHA", U_INSC_AVAGRAHA);
+    INSTALL_ENUM(UIndicSyllabicCategory, "BINDU", U_INSC_BINDU);
+    INSTALL_ENUM(UIndicSyllabicCategory, "BRAHMI_JOINING_NUMBER", U_INSC_BRAHMI_JOINING_NUMBER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CANTILLATION_MARK", U_INSC_CANTILLATION_MARK);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT", U_INSC_CONSONANT);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_DEAD", U_INSC_CONSONANT_DEAD);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_FINAL", U_INSC_CONSONANT_FINAL);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_HEAD_LETTER", U_INSC_CONSONANT_HEAD_LETTER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_INITIAL_POSTFIXED", U_INSC_CONSONANT_INITIAL_POSTFIXED);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_KILLER", U_INSC_CONSONANT_KILLER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_MEDIAL", U_INSC_CONSONANT_MEDIAL);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_PLACEHOLDER", U_INSC_CONSONANT_PLACEHOLDER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_PRECEDING_REPHA", U_INSC_CONSONANT_PRECEDING_REPHA);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_PREFIXED", U_INSC_CONSONANT_PREFIXED);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_SUBJOINED", U_INSC_CONSONANT_SUBJOINED);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_SUCCEEDING_REPHA", U_INSC_CONSONANT_SUCCEEDING_REPHA);
+    INSTALL_ENUM(UIndicSyllabicCategory, "CONSONANT_WITH_STACKER", U_INSC_CONSONANT_WITH_STACKER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "GEMINATION_MARK", U_INSC_GEMINATION_MARK);
+    INSTALL_ENUM(UIndicSyllabicCategory, "INVISIBLE_STACKER", U_INSC_INVISIBLE_STACKER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "JOINER", U_INSC_JOINER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "MODIFYING_LETTER", U_INSC_MODIFYING_LETTER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "NON_JOINER", U_INSC_NON_JOINER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "NUKTA", U_INSC_NUKTA);
+    INSTALL_ENUM(UIndicSyllabicCategory, "NUMBER", U_INSC_NUMBER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "NUMBER_JOINER", U_INSC_NUMBER_JOINER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "PURE_KILLER", U_INSC_PURE_KILLER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "REGISTER_SHIFTER", U_INSC_REGISTER_SHIFTER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "SYLLABLE_MODIFIER", U_INSC_SYLLABLE_MODIFIER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "TONE_LETTER", U_INSC_TONE_LETTER);
+    INSTALL_ENUM(UIndicSyllabicCategory, "TONE_MARK", U_INSC_TONE_MARK);
+    INSTALL_ENUM(UIndicSyllabicCategory, "VIRAMA", U_INSC_VIRAMA);
+    INSTALL_ENUM(UIndicSyllabicCategory, "VISARGA", U_INSC_VISARGA);
+    INSTALL_ENUM(UIndicSyllabicCategory, "VOWEL", U_INSC_VOWEL);
+    INSTALL_ENUM(UIndicSyllabicCategory, "VOWEL_DEPENDENT", U_INSC_VOWEL_DEPENDENT);
+    INSTALL_ENUM(UIndicSyllabicCategory, "VOWEL_INDEPENDENT", U_INSC_VOWEL_INDEPENDENT);
+
+    INSTALL_ENUM(UVerticalOrientation, "ROTATED", U_VO_ROTATED);
+    INSTALL_ENUM(UVerticalOrientation, "TRANSFORMED_ROTATED", U_VO_TRANSFORMED_ROTATED);
+    INSTALL_ENUM(UVerticalOrientation, "TRANSFORMED_UPRIGHT", U_VO_TRANSFORMED_UPRIGHT);
+    INSTALL_ENUM(UVerticalOrientation, "UPRIGHT", U_VO_UPRIGHT);
+#endif
+
 #if U_ICU_VERSION_HEX >= 0x04000000
     INSTALL_ENUM(UWordBreakValues, "CR", U_WB_CR);
     INSTALL_ENUM(UWordBreakValues, "EXTEND", U_WB_EXTEND);
@@ -1477,6 +1650,38 @@ void _init_char(PyObject *m)
     INSTALL_ENUM(ULineBreak, "H3", U_LB_H3);
     INSTALL_ENUM(ULineBreak, "JL", U_LB_JL);
     INSTALL_ENUM(ULineBreak, "JT", U_LB_JT);
+
+    INSTALL_ENUM(UGraphemeClusterBreak, "OTHER", U_GCB_OTHER);
+    INSTALL_ENUM(UGraphemeClusterBreak, "CONTROL", U_GCB_CONTROL);
+    INSTALL_ENUM(UGraphemeClusterBreak, "CR", U_GCB_CR);
+    INSTALL_ENUM(UGraphemeClusterBreak, "EXTEND", U_GCB_EXTEND);
+    INSTALL_ENUM(UGraphemeClusterBreak, "L", U_GCB_L);
+    INSTALL_ENUM(UGraphemeClusterBreak, "LF", U_GCB_LF);
+    INSTALL_ENUM(UGraphemeClusterBreak, "LV", U_GCB_LV);
+    INSTALL_ENUM(UGraphemeClusterBreak, "LVT", U_GCB_LVT);
+    INSTALL_ENUM(UGraphemeClusterBreak, "T", U_GCB_T);
+    INSTALL_ENUM(UGraphemeClusterBreak, "V", U_GCB_V);
+#if U_ICU_VERSION_HEX >= 0x04000000
+    INSTALL_ENUM(UGraphemeClusterBreak, "SPACING_MARK", U_GCB_SPACING_MARK);
+    INSTALL_ENUM(UGraphemeClusterBreak, "PREPEND", U_GCB_PREPEND);
+#endif
+#if U_ICU_VERSION_HEX >= VERSION_HEX(50, 0, 0)
+    INSTALL_ENUM(UGraphemeClusterBreak, "REGIONAL_INDICATOR", U_GCB_REGIONAL_INDICATOR);
+#endif
+#if U_ICU_VERSION_HEX >= VERSION_HEX(58, 0, 0)
+    INSTALL_ENUM(UGraphemeClusterBreak, "E_BASE", U_GCB_E_BASE);
+    INSTALL_ENUM(UGraphemeClusterBreak, "E_BASE_GAZ", U_GCB_E_BASE_GAZ);
+    INSTALL_ENUM(UGraphemeClusterBreak, "GLUE_AFTER_ZWJ", U_GCB_GLUE_AFTER_ZWJ);
+    INSTALL_ENUM(UGraphemeClusterBreak, "ZWJ", U_GCB_ZWJ);
+#endif
+
+    INSTALL_ENUM(UHangulSyllableType, "NOT_APPLICABLE", U_HST_NOT_APPLICABLE);
+    INSTALL_ENUM(UHangulSyllableType, "LEADING_JAMO", U_HST_LEADING_JAMO);
+    INSTALL_ENUM(UHangulSyllableType, "VOWEL_JAMO", U_HST_VOWEL_JAMO);
+    INSTALL_ENUM(UHangulSyllableType, "TRAILING_JAMO", U_HST_TRAILING_JAMO);
+    INSTALL_ENUM(UHangulSyllableType, "LV_SYLLABLE", U_HST_LV_SYLLABLE);
+    INSTALL_ENUM(UHangulSyllableType, "LVT_SYLLABLE", U_HST_LVT_SYLLABLE);
+
 #if U_ICU_VERSION_HEX >= 0x04040000
     INSTALL_ENUM(ULineBreak, "CLOSE_PARENTHESIS", U_LB_CLOSE_PARENTHESIS);
 #endif
